<?php

namespace Drupal\commerce_gmo_linktypeplus\EventSubscriber;

use Drupal\commerce_gmo_linktypeplus\Event\LinkTypePlusEvent;
use Drupal\commerce_order\Entity\Order;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Orderstatus change event subscriber.
 */
class LinkTypePlusEventSubscriber implements EventSubscriberInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Logger Factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * Constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $loggerFactory
   *   Logger .
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, LoggerChannelFactoryInterface $loggerFactory) {
    $this->entityTypeManager = $entityTypeManager;
    $this->loggerFactory = $loggerFactory->get('linktype_event_subscriber');
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    return [
      'Credit' => 'onlinkTypePaymentEvent',
      'PayPay' => 'onlinkTypePaymentEvent',
      'CVS' => 'onlinkTypePaymentEvent',
      'PayEasy' => 'onlinkTypePaymentEvent',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function onlinkTypePaymentEvent(LinkTypePlusEvent $event) {

    $paymentMethod = $event->getPaymentMethod();
    // Get the payment status and update the payment in Drupal.
    $orderId = $event->getOrderId();
    // Remove version from orderId.
    $orderStripId = explode("-", $orderId)[0];
    $remoteState = $event->getTransitionState();
    $drupalStatus = $this->webhookStatusMapper($remoteState);
    $remoteId = $event->getRemoteId();
    $this->loggerFactory->notice("$paymentMethod payment event has been subscribed");
    if (!empty($orderStripId) && !empty($drupalStatus)) {
      if ($this->updatePaymentStatus($orderStripId, $drupalStatus, $remoteId, $remoteState, $paymentMethod, $event)) {
        $this->loggerFactory->notice("The Order " . $orderStripId . ' is updated with the status: ' . $drupalStatus);
        return TRUE;
      }
    }
  }

  /**
   * Update the status in Drupal.
   *
   * @param int $order_id
   *   The Order ID.
   * @param string $status
   *   The Drupal status to be updated.
   * @param string $remote_id
   *   The GMO remote id to be updated.
   * @param string $remote_state
   *   The GMO remote state.
   * @param string $payment_method
   *   The GMO Payment Method.
   * @param Drupal\commerce_gmo_linktypeplus\Event\LinkTypePlusEvent $data
   *   The GMO api response data.
   */
  public function updatePaymentStatus($order_id, $status, $remote_id, $remote_state, $payment_method, LinkTypePlusEvent $data) {
    if ($order_id && !empty($order_id)) {
      $order = Order::load($order_id);
      if (!$order) {
        return TRUE;
      }
      $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
      if (!$order->get('payment_gateway')->entity) {
        return TRUE;
      }
      $paymentGateway = $order->get('payment_gateway')->entity->id();
      $total_price = $order->getTotalprice()->getNumber();
      $currency = $order->getTotalprice()->getCurrencyCode();
      $payment = $payment_storage->loadByProperties([
        'order_id' => $order_id,
      ]);
      if ($status != 'new') {
        if ($payment) {
          if (is_array($payment)) {
            $payment = array_shift($payment);
          }
          $payment->setState($status);
          $payment->setRemoteId($remote_id);
          $payment->setRemoteState($remote_state);
          $payment->setCompletedTime(time());
          $payment->save();
        }
        else {
          $payment = $payment_storage->create([
            'state' => $status,
            'payment_gateway' => $paymentGateway,
            'remote_id' => $remote_id,
            'remote_state' => $remote_state,
            'amount' => [
              'number' => $total_price,
              'currency_code' => $currency,
            ],
            'order_id' => $order_id,
            'completed' => time(),
          ]);
          $payment->save();
        }
      }
      // Apply the transition.
      if ($status == 'completed' && $order->getState()->getId() != 'completed') {
        $order->unlock();
        $order->getState()->applyTransitionById('place');
        $order->setData('GMO_RESPONSE', $data);
        $order->setData('PAYMENT_METHOD', $payment_method);
        $order->setPlacedTime(time());
        $order->save();
        return TRUE;
      }
      if ($status == 'authorization_expired') {
        $order->unlock();
        if ($order->getState()->getId() != 'canceled') {
          $order->getState()->applyTransitionById('cancel');
        }
        $order->setOrderNumber($order->id());
        $order->setData('GMO_RESPONSE', $data);
        $order->setData('PAYMENT_METHOD', $payment_method);
        $order->setPlacedTime(time());
        $order->save();
        return TRUE;
      }
      $order->setOrderNumber($order->id());
      $order->setData('GMO_RESPONSE', $data);
      $order->setData('PAYMENT_METHOD', $payment_method);
      $order->setPlacedTime(time());
      $order->save();
      return TRUE;
    }
  }

  /**
   * Method used to track the linktype status with the drupal commerce_payment.
   *
   * We got these status through webhook.
   *
   * It may differs for credit card, paypay etc.
   *
   * @todo Please add the required status by refering the doc
   *
   * Refer: https://docs.mul-pay.jp/payment/credit/notice
   *        https://docs.mul-pay.jp/paypay/payg-notice
   *        https://docs.mul-pay.jp/
   */
  public function webhookStatusMapper($state) {
    switch ($state) {
      case 'SAUTH':
      case 'AUTH':
      case 'REQSUCCESS':
      case 'AUTHPROCESS':
        return 'authorization';

      case 'PAYSTART':
        return 'new';

      case 'ERROR':
        return 'authorization_expired';

      case 'PAYSUCCESS':
      case 'SALES':
      case 'TRADING':
        return 'completed';

      case 'UNPROCESSED':
      case 'AUTHENTICATED':
      case 'CAPTURE':
        return 'new';

      case 'CANCEL':
        return 'authorization_expired';

      case 'PAYFAIL':
      case 'VOID':

        return 'authorization_expired';

      default:
        return 'new';
    }

  }

}
